import numpy as np
from osgeo import gdal
from .raster import RasterInterpolator
from scipy.ndimage.morphology import distance_transform_edt
from math import ceil

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright Jon Hill, University of York, jon.hill@york.ac.uk

# read in a raster and create the buffer from 0 at edge, to 1 at distance


class CreateBuffer():
    """
    Implements the creation of a distance buffer from the edge of
    a raster to the centre::

        rbuff = CreateBuffer('myRaster.tif',10000.0)

    Will create a buffer raster with the same extents as myRaster.tif
    with a buffer that goes from 0 at the edge to 1.0 at a distance of
    10,000 units from the edge. The distance should be in the same
    units as the raster file.

    You can also specify the "resolution" of your buffer using the 'over'
    argument. Using '10' would use ten units resolve the buffer from
    edge to distance (e.g. is distance was 1.5, and over was 10, your output
    buffer would have a dx of 0.15). This will probably alter the extents
    of your buffer raster such that it no longer matches the actual raster,
    so proceed with caution. It may however, be useful if your input raster
    has very high resolution and you want to prevent multiple large raster
    files.

    Once the object is made, write out the buffer using::

        rbuff.make_buffer('output_buffer.tif')

    Any GDAL-understood file format is supported for input or output.
    """

    def __init__(self, filename, distance, over=None):
        """
        Init our buffer

        Args:
            filename: filename to write to
            distance: distance (in raster units) to extend buffer over
            over: alter the distance to be in some other units

        Returns:
            a createBuffer object
        """
        self.distance = distance
        self.over = over
        self.raster = RasterInterpolator(filename)
        self.raster.set_band()
        self.extent = self.raster.get_extent()

    def __write_raster__(self, filename, array, dx, origin, proj):
        """
        Write raster to file

        Args:
            filename: filename to write to
            array: the data. 2D numpy array
            dx: resolution (length 2 list)
            origin: the LLC coordinates
            proj: Projection space for the raster (wkt)

        Returns:
            Nothing
        """
        dst_filename = filename
        x_pixels = array.shape[1]
        y_pixels = array.shape[0]
        x_min = origin[0][0]
        y_max = origin[0][1] + dx[1]*y_pixels
        wkt_projection = proj

        driver = gdal.GetDriverByName('GTiff')
        dataset = driver.Create(
            dst_filename,
            x_pixels,
            y_pixels,
            1,
            gdal.GDT_Float32)

        dataset.SetGeoTransform((
            x_min,       # 0
            dx[0],       # 1
            0,           # 2
            y_max,       # 3
            0,           # 4
            -dx[1]))

        dataset.SetProjection(wkt_projection)
        dataset.GetRasterBand(1).WriteArray(array)
        dataset.FlushCache()
        return

    def extend_mask(self, array, iterations):
        """
        Extend the mask a number of "cells" in all directions"

        Args:
            array: the numpy aray to extend
            iterations: integer for how many cells to extend

        Returns:
            a numpy array
        """

        # function to extend a mask array.
        # Taken from: http://www.siafoo.net/snippet/82
        # Copyright 2007 Regents University of California
        # Written by David Isaacson under BSD licence
        yLen, xLen = array.shape
        output = array.copy()
        for i in range(iterations):
            for y in range(yLen):
                for x in range(xLen):
                    if (y > 0 and array[y-1, x]) or \
                       (y < yLen - 1 and array[y+1, x]) or \
                       (x > 0 and array[y, x-1]) or \
                       (x < xLen - 1 and array[y, x+1]):
                        output[y, x] = True
            array = output.copy()

        return output

    def make_buffer(self, output_file):
        """
        Create a buffer raster from 0 to 1 over a set distance.

        Args:
            output_file: where to save this raster
        """

        # make a raster of the same extent, but with
        # square resolution which is dependant on distance buffer

        llc = self.extent[1]
        urc = self.extent[3]
        if self.over is None:
            transform = self.raster.ds.GetGeoTransform()
            dx = [transform[1], -transform[5]]
            nrows = self.raster.ds.RasterXSize
            ncols = self.raster.ds.RasterYSize
        else:
            dx = [self.distance / self.over, self.distance / self.over]
            # note this changes our extent if "over" is set
            nrows = int(ceil((urc[0] - llc[0]) / dx[0]))
            ncols = int(ceil((urc[1] - llc[1]) / dx[1]))

        # fill with edge value
        dist = np.full((ncols, nrows), 0.0)
        # then fill in the middle,
        # except where there is no data
        dist[1:-1, 1:-1] = 1
        if self.over is None:
            orig_raster = self.raster.get_array()
            nodata = self.raster.nodata
            # TODO this will only work if dist and orig_raster
            # are the same size
            dist[orig_raster == nodata] = 0
            # they also be nan...(shouldn't as we swap it out...)
            dist[np.isnan(orig_raster)] = 0
            # we now extend this mask - we only need to do this, if the
            # no data occurs (i.e. no contiguous data)
            if (nodata in orig_raster):
                mask = np.full((ncols, nrows), False)
                mask[dist == 0] = True
                mask = self.extend_mask(mask, 1)
                dist[mask] = 0
        # calc euclidian distance and convert to 0 -> 1 scale
        dist = distance_transform_edt(dist, sampling=[dx[0], dx[1]])
        dist = dist / self.distance
        dist[dist > 1] = 1.0

        # create a suitable output filename
        self.__write_raster__(output_file, np.flipud(dist), dx, [llc, urc],
                              self.raster.ds.GetProjection())



import numpy as np
from osgeo import gdal
import math

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright Jon Hill, University of York, jon.hill@york.ac.uk

"""
This module contains the Interpolator class and the RasterInterpolator
class. These RasterInterpolator class reads in a GDAL raster and
then allows interagation of that data at an arbitrary point using
bi-linear interpolation.
"""


class RasterInterpolatorError(Exception):
    """
    A generic error created by the RasterInterpolator class
    """

    pass


class CoordinateError(RasterInterpolatorError):
    """
        Raised when a point is outside the raster or the land mask
    """
    def __init__(self, message, x, i, j):
        self.message = message
        self.x = x
        self.ij = i, j

    def __str__(self):
        return "at x, y={} indexed at i, j={}; {}".format(self.x, self.ij,
                                                          self.message)


class Interpolator():
    """Implements an object to interpolate values from a Raster-type data set.

    Used by the RasterInterpolator. A separate object as in the future we
    may switch bands and hence have to reload the val data.
    """

    def __init__(self, origin, delta, val, mask=None, minmax=None):
        """
        Init our Interpolator

        Args:
            origin: where the raster starts (LLC)
            delta: resolution (length 2 list [dx,dy])
            val: 2D numpy array of data
            mask: a land mask (not yet implemented)
            minmax: any min/max values to adhere to (length 2 list [min,max])

        Returns:
            a Interpolator object
        """
        self.origin = origin
        self.delta = delta
        self.val = val
        self.mask = mask
        self.minmax = minmax

    def set_mask(self, mask):
        """
        Set a mask to use. Not yet implemented.
        """
        self.mask = mask

    def get_val(self, point):
        """
        Get the value of this raster at the desired point via bi-linear
        interpolation.

        Args:
            point: a length 2 list containing x,y coordinates

        Returns:
            The value of the raster stack at that point

        Raises:
            CoordinateError: The point is outside the rasters
            RasterInterpolatorError: Generic error interpolating data at
                that point
        """

        yhat = ((point[0]+(self.delta[0]/2.0)-self.origin[0])/self.delta[0])
        xhat = ((point[1]+(self.delta[1]/2.0)-self.origin[1])/self.delta[1])
        j = int(math.floor(yhat))-1
        i = int(math.floor(xhat))-1
        # this is not caught as an IndexError below, because of wrapping of
        # negative indices
        if i < 0 or j < 0:
            raise CoordinateError("Coordinate out of range", point, i, j)
        alpha = (xhat) % 1.0
        beta = (yhat) % 1.0
        neigh_i = i+1
        neigh_j = j+1
        if neigh_i < 0 or neigh_j < 0:
            raise CoordinateError("Coordinate out of range", point, i, j)
        try:
            if self.mask is not None:
                # case with a land mask - masks not yet implemented!
                w00 = (1.0-alpha)*(1.0-beta)*self.mask[i, j]
                w10 = alpha*(1.0-beta)*self.mask[i+1, j]
                w01 = (1.0-alpha)*beta*self.mask[i, j+1]
                w11 = alpha*beta*self.mask[i+1, j+1]
                if len(self.val.shape) == 2:
                    value = w00*self.val[i, j] + w10*self.val[i+1, j] \
                            + w01*self.val[i, j+1] + w11*self.val[i+1, j+1]
                else:
                    raise RasterInterpolatorError("Field to interpolate,"
                                                  "should have 2 dimensions")
                sumw = w00+w10+w01+w11

                if sumw > 0.0:
                    value = value/sumw
                else:
                    raise CoordinateError("Probing point inside land mask",
                                          point, i, j)

            else:
                # case without a land mask
                if len(self.val.shape) == 2:
                    value = ((1.0-beta)*((1.0-alpha)*self.val[i, j] +
                                         alpha*self.val[neigh_i, j]) +
                             beta*((1.0-alpha)*self.val[i, neigh_j] +
                                   alpha*self.val[neigh_i, neigh_j]))
                else:
                    raise RasterInterpolatorError("Field to interpolate,"
                                                  "should have 2 dimensions")
        except IndexError:
            raise CoordinateError("Coordinate out of range", point, i, j)

        if self.minmax is not None:
            if self.minmax[0] is not None:
                if value < self.minmax[0]:
                    value = self.minmax[0]
            if self.minmax[1] is not None:
                if value > self.minmax[1]:
                    value = self.minmax[1]

        return value


# note that a RasterInterpolator is *not* object an Interpolator object
# the latter is considered immutable, whereas the NetCDFInterpolator may
# change in future
class RasterInterpolator(object):
    """
    Implements an object to interpolate values from a Raster-stored data set::

        rci = RasterInterpolator('foo.tif')

    Any GDAL supported raster format should be fine. The origin is assumed
    to be the lower-left corner (i.e. south west)
    and the projection space is stored with the raster.

    To indicate the band to be interpolated::

        rci.set_band(2)

    The default is Band 1 (ie. with no number given)

    To interpolate this field in any arbitrary point::

        rci.get_val((-3.0, 58.5))

    It is allowed to switch between different fields using multiple
    calls of set_band().

    """
    def __init__(self, filename, minmax=None):
        """
        Init our RasterInterpolator

        Args:
            filename: Which raster to load
            minmax: any min/max values to adhere to (length 2 list [min,max])

        Returns:
            a RasterInterpolator object
        """
        self.ds = gdal.Open(filename)
        if (self.ds is None):
            raise RasterInterpolatorError("Couldn't find your raster file:" +
                                          filename + ". Exiting.")
        self.band = None
        self.mask = None
        self.interpolator = None
        self.extent = None
        self.dx = 0.0
        self.nodata = None
        self.minmax = minmax

    def get_extent(self):
        """Return list of corner coordinates from a geotransform

        Returns:
            List continaing the corner coordinates of the raster

        """
        cols = self.ds.RasterXSize
        rows = self.ds.RasterYSize
        gt = self.ds.GetGeoTransform()

        ext = []
        xarr = [0, cols]
        yarr = [0, rows]

        for px in xarr:
            for py in yarr:
                x = gt[0] + (px * gt[1]) + (py * gt[2])
                y = gt[3] + (px * gt[4]) + (py * gt[5])
                ext.append([x, y])
            yarr.reverse()
        return ext

    def set_band(self, band_no=1):
        """
        Set the number of the band to be used. Usually 1, which is default

        Args:
            band_no: an integer which is the band number to use. Default is 1.

        """
        self.band = band_no
        raster = self.ds.GetRasterBand(self.band)
        self.nodata = raster.GetNoDataValue()
        self.val = np.flipud(np.array(raster.ReadAsArray()))
        # fix any NAN with the no-data value
        if (np.isnan(self.val).any()):
            self.val[np.isnan(self.val)] = self.nodata
        self.extent = self.get_extent()
        origin = np.amin(self.extent, axis=0)
        transform = self.ds.GetGeoTransform()
        self.dx = [transform[1], -transform[5]]
        self.interpolator = Interpolator(origin, self.dx, self.val,
                                         self.mask, self.minmax)

    def get_array(self):
        """
        Get the raw data in the raster

        Returns:
            a numpy array containing the raster data
        """
        if (self.interpolator is None):
            raise RasterInterpolatorError("Should call set_band() "
                                          "before calling get_array()!")
        return self.val

    def get_val(self, x):
        """
        Interpolate the field chosen with set_field().
        The order of the coordinates should correspond
        with the storage order in the file.

        Args:
            point: a length 2 list containing x,y coordinates

        Returns:
            The value of the raster stack at that point

        Raises:
            CoordinateError: The point is outside the rasters
            RasterInterpolatorError: Generic error interpolating
                data at that point

        """
        if (self.interpolator is None):
            raise RasterInterpolatorError("Should call set_band() "
                                          "before calling get_val()!")
        val = self.interpolator.get_val(x)
        return val

    def point_in(self, point):
        """
        Does a point lay inside a raster's extent?

        Args:
            point: a length 2 list containing x,y coordinates

        Returns:
            Boolean. True if point is in the raster. False otherwise.
        """

        # does this point occur in the raster?
        llc = np.amin(self.extent, axis=0)+(self.dx[0]/2)
        urc = np.amax(self.extent, axis=0)-(self.dx[1]/2)
        if ((point[0] <= urc[0] and point[0] >= llc[0]) and
           (point[1] <= urc[1] and point[1] >= llc[1])):
            return True
        else:
            return False



from .raster import RasterInterpolator
from .raster_buffer import CreateBuffer
import os
from shutil import copyfile
import tempfile
try:
    from itertools import izip as zip
except ImportError:  # will be 3.x series
    pass

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright Jon Hill, University of York, jon.hill@york.ac.uk

# create a hierarchy of rasters to pull data from, smoothly blending
# between them


class HRDSError(Exception):
    # any error generated by the RasterInterpolator object:

    pass


class HRDS():
    """
    The main HRDS class. Create a raster stack and initialise::

        bathy = HRDS("gebco_uk.tif",
             rasters=("emod_utm.tif",
                      "marine_digimap.tif"),
             distances=(10000, 5000),
             minmax=None)
        bathy.set_bands()

    The first argument is the base raster filename. `rasters` is a list
    of raster filenames, with corresponding `distances` over which to
    create the buffer. Distances are in the same units as the rasters.
    The min/max argument allows you to specify a minimum or maximum
    (or both!) values when returning data. This is useful for ocean
    simulations where you want a minimum depth to prevent "drying".
    To set this, do::

        bathy = HRDS("gebco_uk.tif",
             rasters=("emod_utm.tif",
                      "marine_digimap.tif"),
             distances=(10000, 5000),
             minmax=[[None,-5],[None,-3],[None,None]])

    which would set a maximum depth of -5m on the gebco data (-ve = below
    sea level, +ve above), maximum of -3m on the emod data and no limits
    on the marine_digimap data. You must supply the same number of min-max
    pairs are there are total rasters. Temporary buffer files will be created
    and deleted at clean up

    If is possible to supply the buffer rasters directly (e.g. if you want
    to use different distances on each edge of your raster, or some other
    such thing). Buffer extent must match or exceed the corresponding raster
    extent::

        bathy = HRDS("gebco_uk.tif",
             rasters=("emod_utm.tif",
                      "marine_digimap.tif"),
             buffers=(buffer1.tif,
                      buffer2.tif,
                      buffer3.tif))
        bathy.set_bands()

    Once set up, you can ask for data at any point::

        bathy.get_val(100,100)
    """
    def __init__(self, baseRaster, rasters=None, distances=None,
                 buffers=None, minmax=None, saveBuffers=False):
        """
        Set up our hrds object

        Args:
          baseRaster: the low res raster filename across whole domain.
          rasters: a list of filenames of the other rasters in priority order.
          distances: the distance to create a buffer (in same units as
            corresponding raster) for each.
          buffers: a list of buffer filenames in the same order as rasters
            if created already. In this case, don't supply distances.
          minmax: list of minimum and maximum values to use
          saveBuffers: boolean to save buffers if needed
        """

        if distances is None:
            if len(rasters) != len(buffers):
                raise HRDSError("You have "+str(len(rasters)) +
                                "rasters and "+str(len(buffers)) +
                                "buffers. They should match")
        else:
            if len(rasters) != len(distances):
                raise HRDSError("You have "+str(len(rasters)) +
                                "rasters and "+str(len(distances)) +
                                "distances. They should match")

        if minmax is not None:
            if len(rasters)+1 != len(minmax):
                raise HRDSError("Please supply same number of minmax values" +
                                "as the total number of rasters, inc. base." +
                                "You gave me: "+str(len(minmax))+" min/max" +
                                "and I expected: "+str(len(rasters)+1))

        if minmax is None:
            self.baseRaster = RasterInterpolator(baseRaster)
        else:
            self.baseRaster = RasterInterpolator(baseRaster, minmax[0])
        self.raster_stack = []
        for i, r in enumerate(rasters):
            if minmax is None:
                self.raster_stack.append(RasterInterpolator(r))
            else:
                self.raster_stack.append(RasterInterpolator(r, minmax[i+1]))
        self.buffer_stack = []
        # the user is asking us to create the buffer files
        if buffers is None:
            # we create the files in a temp dir and if the user wants
            # them afterwards we copy to a sensible name
            with tempfile.TemporaryDirectory() as tmpdirname:
                for r, d in zip(rasters, distances):
                    # create buffer file name, based on raster filename
                    keep_buf_file = os.path.splitext(r)[0]+"_buffer.tif"
                    temp_buf_file = os.path.join(tmpdirname,
                                                 os.path.splitext(
                                                     os.path.basename(r))[0] +
                                                 "_buffer.tif")
                    # create buffer
                    rbuff = CreateBuffer(r, d)
                    rbuff.make_buffer(temp_buf_file)
                    # add to stack and store in memory
                    self.buffer_stack.append(RasterInterpolator(temp_buf_file))
                    # does the user also want the file saving?
                    if saveBuffers:
                        copyfile(temp_buf_file, keep_buf_file)

        else:
            # create buffer stack from filenames
            for r in buffers:
                self.buffer_stack.append(RasterInterpolator(r))

        # reverse the arrays
        self.buffer_stack.reverse()
        self.raster_stack.reverse()

    def set_bands(self, bands=None):
        """
        Performs bilinear interpolation of your raster stack
        to give a value at the requested point.

        Args:
            bands: a list of band numbers for each raster in the stack or None
                (uses the first band in each raster). Default is None.

        """

        if bands is None:
            self.baseRaster.set_band()
            for r in self.raster_stack:
                r.set_band()
            for r in self.buffer_stack:
                r.set_band()
        else:
            counter = 1
            self.baseRaster.bands(bands[0])
            for r in self.raster_stack:
                r.set_band(bands[counter])
                counter += 1
            counter = 1
            for r in self.buffer_stack:
                r.set_band(bands[counter])
                counter += 1

    def get_val(self, point):
        """
        Performs bilinear interpolation of your raster stack
        to give a value at the requested point.

        Args:
            point: a length 2 list containing x,y coordinates

        Returns:
            The value of the raster stack at that point

        Raises:
            CoordinateError: The point is outside the rasters
            RasterInterpolatorError: Generic error interpolating
                data at that point
        """
        # determine if we're in any of the rasters in the list,
        # starting from the last one
        for i, r, b in zip(range(0, len(self.raster_stack)+1),
                           self.raster_stack,
                           self.buffer_stack):
            if r.point_in(point):
                # if so, check the buffer value
                if b.get_val(point) == 1.0:
                    return r.get_val(point)
                else:
                    for rr in self.raster_stack[i+1:]:
                        # if not, find the next raster we're in, inc. the base
                        if (rr.point_in(point) and
                           not self.buffer_stack[i+1].get_val(point) == 0):
                            val = r.get_val(point)*b.get_val(point) + \
                                  rr.get_val(point)*(1-b.get_val(point))
                            return val
                    # if we get here, there is no other layer,
                    # so use base raster
                    val = r.get_val(point)*b.get_val(point) + \
                        self.baseRaster.get_val(point)*(1-b.get_val(point))
                    return val

        # we're not in the raster stack, so return value from base
        return self.baseRaster.get_val(point)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright Jon Hill, University of York, jon.hill@york.ac.uk

"""
hrds contains the following components:

    * buffer - functions to create buffer files
    * hdrs - the main hdrs object.
    * RasterInterpolator - objects to interpolate individual rasters

"""

__version__ = "0.1.3"

from .raster import RasterInterpolator   # NOQA
from .raster_buffer import CreateBuffer  # NOQA
from .hrds import HRDS                   # NOQA



import setuptools

DESCRIPTION = "hrds is a python package for obtaining points from a set of rasters at different resolutions. You can request a point (or list of points) and hrds will return a value based on the highest resolution dataset (as defined by the user) available at that point, blending datasets in a buffer region to ensure consistancy."


def get_version():
    """
    kludgy, but lets you set version in one place, and have it in the package
    """
    with open("hrds/__init__.py") as initfile:
        for line in initfile:
            parts = line.strip().split("=")
            if parts[0].strip() == "__version__":
                version = parts[1].strip().strip("'").strip('"')
                return version
    raise ValueError("no __version__ defined in package __init__")


setuptools.setup(name='hrds',
                 version=get_version(),
                 author='Jon Hill',
                 author_email='jon.hill@york.ac.uk',
                 description=DESCRIPTION,
                 url='https://github.com/EnvModellingGroup/hrds',
                 packages=setuptools.find_packages(),
                 keywords=['raster', 'gis', 'modelling'],
                 classifiers=[
                   'Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
                   'Programming Language :: Python :: 3.6',
                   'Topic :: Scientific/Engineering'],
                 )
